/*
 * Copyright Elasticsearch B.V. and/or licensed to Elasticsearch B.V. under one
 * or more contributor license agreements. Licensed under the Elastic License
 * 2.0; you may not use this file except in compliance with the Elastic License
 * 2.0.
 */
package org.elasticsearch.xpack.core.scheduler;

import org.elasticsearch.ElasticsearchParseException;
import org.elasticsearch.common.time.DateFormatter;
import org.elasticsearch.core.SuppressForbidden;
import org.elasticsearch.xcontent.ToXContentFragment;
import org.elasticsearch.xcontent.XContentBuilder;

import java.io.IOException;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.time.ZoneOffset;
import java.time.temporal.ChronoField;
import java.util.Calendar;
import java.util.Iterator;
import java.util.Locale;
import java.util.Map;
import java.util.Objects;
import java.util.SortedSet;
import java.util.StringTokenizer;
import java.util.TimeZone;
import java.util.TreeSet;

import static java.util.Map.entry;
import static org.elasticsearch.xpack.core.watcher.support.Exceptions.illegalArgument;

/**
 * THIS CLASS IS A FORK OF
 * <a href="https://fisheye.terracotta.org/browse/Quartz/trunk/quartz-core/src/main/java/org/quartz/CronExpression.java?r=2426">
 *     {@code CronExpression}</a>
 * FROM THE <a href="http://quartz-scheduler.org/">QUARTZ</a> PROJECT
 *
 *
 * Provides a parser and evaluator for unix-like cron expressions. Cron
 * expressions provide the ability to specify complex time combinations such as
 * &quot;At 8:00am every Monday through Friday&quot; or &quot;At 1:30am every
 * last Friday of the month&quot;.
 * <P>
 * Cron expressions are comprised of 6 required fields and one optional field
 * separated by white space. The fields respectively are described as follows:
 *
 * <table style="border-collapse: separate; border-spacing: 8px;">
 * <caption>Fields in cron expressions</caption>
 * <tr>
 * <th>Field Name</th>
 * <th>&nbsp;</th>
 * <th>Allowed Values</th>
 * <th>&nbsp;</th>
 * <th>Allowed Special Characters</th>
 * </tr>
 * <tr>
 * <td><code>Seconds</code></td>
 * <td>&nbsp;</td>
 * <td><code>0-59</code></td>
 * <td>&nbsp;</td>
 * <td><code>, - * /</code></td>
 * </tr>
 * <tr>
 * <td><code>Minutes</code></td>
 * <td>&nbsp;</td>
 * <td><code>0-59</code></td>
 * <td>&nbsp;</td>
 * <td><code>, - * /</code></td>
 * </tr>
 * <tr>
 * <td><code>Hours</code></td>
 * <td>&nbsp;</td>
 * <td><code>0-23</code></td>
 * <td>&nbsp;</td>
 * <td><code>, - * /</code></td>
 * </tr>
 * <tr>
 * <td><code>Day-of-month</code></td>
 * <td>&nbsp;</td>
 * <td><code>1-31</code></td>
 * <td>&nbsp;</td>
 * <td><code>, - * ? / L W</code></td>
 * </tr>
 * <tr>
 * <td><code>Month</code></td>
 * <td>&nbsp;</td>
 * <td><code>0-11 or JAN-DEC</code></td>
 * <td>&nbsp;</td>
 * <td><code>, - * /</code></td>
 * </tr>
 * <tr>
 * <td><code>Day-of-Week</code></td>
 * <td>&nbsp;</td>
 * <td><code>1-7 or SUN-SAT</code></td>
 * <td>&nbsp;</td>
 * <td><code>, - * ? / L #</code></td>
 * </tr>
 * <tr>
 * <td><code>Year (Optional)</code></td>
 * <td>&nbsp;</td>
 * <td><code>empty, 1970-2199</code></td>
 * <td>&nbsp;</td>
 * <td><code>, - * /</code></td>
 * </tr>
 * </table>
 * <P>
 * The '*' character is used to specify all values. For example, &quot;*&quot;
 * in the minute field means &quot;every minute&quot;.
 * <P>
 * The '?' character is allowed for the day-of-month and day-of-week fields. It
 * is used to specify 'no specific value'. This is useful when you need to
 * specify something in one of the two fields, but not the other.
 * <P>
 * The '-' character is used to specify ranges For example &quot;10-12&quot; in
 * the hour field means &quot;the hours 10, 11 and 12&quot;.
 * <P>
 * The ',' character is used to specify additional values. For example
 * &quot;MON,WED,FRI&quot; in the day-of-week field means &quot;the days Monday,
 * Wednesday, and Friday&quot;.
 * <P>
 * The '/' character is used to specify increments. For example &quot;0/15&quot;
 * in the seconds field means &quot;the seconds 0, 15, 30, and 45&quot;. And
 * &quot;5/15&quot; in the seconds field means &quot;the seconds 5, 20, 35, and
 * 50&quot;.  Specifying '*' before the  '/' is equivalent to specifying 0 is
 * the value to start with. Essentially, for each field in the expression, there
 * is a set of numbers that can be turned on or off. For seconds and minutes,
 * the numbers range from 0 to 59. For hours 0 to 23, for days of the month 0 to
 * 31, and for months 0 to 11 (JAN to DEC). The &quot;/&quot; character simply helps you turn
 * on every &quot;nth&quot; value in the given set. Thus &quot;7/6&quot; in the
 * month field only turns on month &quot;7&quot;, it does NOT mean every 6th
 * month, please note that subtlety.
 * <P>
 * The 'L' character is allowed for the day-of-month and day-of-week fields.
 * This character is short-hand for &quot;last&quot;, but it has different
 * meaning in each of the two fields. For example, the value &quot;L&quot; in
 * the day-of-month field means &quot;the last day of the month&quot; - day 31
 * for January, day 28 for February on non-leap years. If used in the
 * day-of-week field by itself, it simply means &quot;7&quot; or
 * &quot;SAT&quot;. But if used in the day-of-week field after another value, it
 * means &quot;the last xxx day of the month&quot; - for example &quot;6L&quot;
 * means &quot;the last friday of the month&quot;. You can also specify an offset
 * from the last day of the month, such as "L-3" which would mean the third-to-last
 * day of the calendar month. <i>When using the 'L' option, it is important not to
 * specify lists, or ranges of values, as you'll get confusing/unexpected results.</i>
 * <P>
 * The 'W' character is allowed for the day-of-month field.  This character
 * is used to specify the weekday (Monday-Friday) nearest the given day.  As an
 * example, if you were to specify &quot;15W&quot; as the value for the
 * day-of-month field, the meaning is: &quot;the nearest weekday to the 15th of
 * the month&quot;. So if the 15th is a Saturday, the trigger will fire on
 * Friday the 14th. If the 15th is a Sunday, the trigger will fire on Monday the
 * 16th. If the 15th is a Tuesday, then it will fire on Tuesday the 15th.
 * However if you specify &quot;1W&quot; as the value for day-of-month, and the
 * 1st is a Saturday, the trigger will fire on Monday the 3rd, as it will not
 * 'jump' over the boundary of a month's days.  The 'W' character can only be
 * specified when the day-of-month is a single day, not a range or list of days.
 * <P>
 * The 'L' and 'W' characters can also be combined for the day-of-month
 * expression to yield 'LW', which translates to &quot;last weekday of the
 * month&quot;.
 * <P>
 * The '#' character is allowed for the day-of-week field. This character is
 * used to specify &quot;the nth&quot; XXX day of the month. For example, the
 * value of &quot;6#3&quot; in the day-of-week field means the third Friday of
 * the month (day 6 = Friday and &quot;#3&quot; = the 3rd one in the month).
 * Other examples: &quot;2#1&quot; = the first Monday of the month and
 * &quot;4#5&quot; = the fifth Wednesday of the month. Note that if you specify
 * &quot;#5&quot; and there is not 5 of the given day-of-week in the month, then
 * no firing will occur that month.  If the '#' character is used, there can
 * only be one expression in the day-of-week field (&quot;3#1,6#3&quot; is
 * not valid, since there are two expressions).
 * <!--The 'C' character is allowed for the day-of-month and day-of-week fields.
 * This character is short-hand for "calendar". This means values are
 * calculated against the associated calendar, if any. If no calendar is
 * associated, then it is equivalent to having an all-inclusive calendar. A
 * value of "5C" in the day-of-month field means "the first day included by the
 * calendar on or after the 5th". A value of "1C" in the day-of-week field
 * means "the first day included by the calendar on or after Sunday".-->
 * <P>
 * The legal characters and the names of months and days of the week are not
 * case sensitive.
 *
 * <p>
 * <b>NOTES:</b>
 * <ul>
 * <li>Support for specifying both a day-of-week and a day-of-month value is
 * not complete (you'll need to use the '?' character in one of these fields).
 * </li>
 * <li>Overflowing ranges is supported - that is, having a larger number on
 * the left hand side than the right. You might do 22-2 to catch 10 o'clock
 * at night until 2 o'clock in the morning, or you might have NOV-FEB. It is
 * very important to note that overuse of overflowing ranges creates ranges
 * that don't make sense and no effort has been made to determine which
 * interpretation CronExpression chooses. An example would be
 * "0 0 14-6 ? * FRI-MON". </li>
 * </ul>
 *
 *
 * @author Sharada Jambula, James House
 * @author Contributions from Mads Henderson
 * @author Refactoring from CronTrigger to CronExpression by Aaron Craven
 */
public class Cron implements ToXContentFragment {
    protected static final TimeZone UTC = TimeZone.getTimeZone(ZoneOffset.UTC);
    protected static final DateFormatter formatter = DateFormatter.forPattern("yyyy-MM-dd'T'HH:mm:ss");

    private static final int SECOND = 0;
    private static final int MINUTE = 1;
    private static final int HOUR = 2;
    private static final int DAY_OF_MONTH = 3;
    private static final int MONTH = 4;
    private static final int DAY_OF_WEEK = 5;
    private static final int YEAR = 6;
    private static final int ALL_SPEC_INT = 99; // '*'
    private static final int NO_SPEC_INT = 98; // '?'
    private static final Integer ALL_SPEC = ALL_SPEC_INT;
    private static final Integer NO_SPEC = NO_SPEC_INT;

    private static final Map<String, Integer> MONTH_MAP = Map.ofEntries(
        entry("JAN", 0),
        entry("FEB", 1),
        entry("MAR", 2),
        entry("APR", 3),
        entry("MAY", 4),
        entry("JUN", 5),
        entry("JUL", 6),
        entry("AUG", 7),
        entry("SEP", 8),
        entry("OCT", 9),
        entry("NOV", 10),
        entry("DEC", 11)
    );

    private static final Map<String, Integer> DAY_MAP = Map.of("SUN", 1, "MON", 2, "TUE", 3, "WED", 4, "THU", 5, "FRI", 6, "SAT", 7);

    private final String expression;

    private ZoneId timeZone;

    private transient TreeSet<Integer> seconds;
    private transient TreeSet<Integer> minutes;
    private transient TreeSet<Integer> hours;
    private transient TreeSet<Integer> daysOfMonth;
    private transient TreeSet<Integer> months;
    private transient TreeSet<Integer> daysOfWeek;
    private transient TreeSet<Integer> years;

    private transient boolean lastdayOfWeek = false;
    private transient int nthdayOfWeek = 0;
    private transient boolean lastdayOfMonth = false;
    private transient boolean nearestWeekday = false;
    private transient int lastdayOffset = 0;

    // Restricted to 50 years as the tzdb only has correct DST transition information for countries using a lunar calendar
    // for the next ~60 years
    public static final int MAX_YEAR = Calendar.getInstance(UTC, Locale.ROOT).get(Calendar.YEAR) + 50;

    public Cron(String expression, ZoneId timeZone) {
        this.timeZone = timeZone;
        assert expression != null : "cron expression cannot be null";
        this.expression = expression.toUpperCase(Locale.ROOT);
        try {
            buildExpression(this.expression);
        } catch (Exception e) {
            throw illegalArgument("invalid cron expression [{}]", e, expression);
        }
    }

    /**
     * Constructs a new <CODE>CronExpression</CODE> based on the specified
     * parameter.
     *
     * @param expression String representation of the cron expression the
     *                       new object should represent
     * @throws IllegalArgumentException
     *         if the string expression cannot be parsed into a valid
     *         <CODE>CronExpression</CODE>
     */
    public Cron(String expression) {
        this(expression, UTC.toZoneId());
    }

    /**
     * Constructs a new {@code CronExpression} as a copy of an existing
     * instance.
     *
     * @param cron The existing cron expression to be copied
     */
    public Cron(Cron cron) {
        this(cron.expression, cron.timeZone);
    }

    public void setTimeZone(ZoneId timeZone) {
        this.timeZone = timeZone;
    }

    /**
     * Returns the next date/time <I>after</I> the given date/time which
     * satisfies the cron expression.
     *
     * @param time the time since the epoch, or -1 if next time is unsupported (e.g. the cron expression points to
     *             a time that is previous to the given time)
     * @return the next valid time (since the epoch)
     */
    @SuppressForbidden(reason = "In this case, the DST ambiguity of the atZone method is desired, understood and tested")
    public long getNextValidTimeAfter(final long time) {

        LocalDateTime afterTimeLdt = LocalDateTime.ofInstant(java.time.Instant.ofEpochMilli(time), timeZone).plusSeconds(1);
        LocalDateTimeLegacyWrapper cl = new LocalDateTimeLegacyWrapper(afterTimeLdt.with(ChronoField.MILLI_OF_SECOND, 0));

        boolean gotOne = false;
        // loop until we've computed the next time, or we've past the endTime
        while (gotOne == false) {

            if (cl.getYear() > 2999) { // prevent endless loop...
                return -1;
            }

            SortedSet<Integer> st = null;
            int t = 0;

            int sec = cl.getSecond();
            int min = cl.getMinute();

            // get second.................................................
            st = seconds.tailSet(sec);
            if (st != null && st.size() != 0) {
                sec = st.first();
            } else {
                sec = seconds.first();
                min++;
                cl.setMinute(min);
            }
            cl.setSecond(sec);

            min = cl.getMinute();
            int hr = cl.getHour();
            t = -1;

            // get minute.................................................
            st = minutes.tailSet(min);
            if (st != null && st.size() != 0) {
                t = min;
                min = st.first();
            } else {
                min = minutes.first();
                hr++;
            }
            if (min != t) {
                cl.setSecond(0);
                cl.setMinute(min);
                cl.setHour(hr);
                continue;
            }
            cl.setMinute(min);

            hr = cl.getHour();
            int day = cl.getDayOfMonth();
            t = -1;

            // get hour...................................................
            st = hours.tailSet(hr);
            if (st != null && st.size() != 0) {
                t = hr;
                hr = st.first();
            } else {
                hr = hours.first();
                day++;
            }
            if (hr != t) {
                cl.setSecond(0);
                cl.setMinute(0);
                cl.setDayOfMonth(day);
                cl.setHour(hr);
                continue;
            }
            cl.setHour(hr);

            day = cl.getDayOfMonth();
            int mon = cl.getMonth() + 1;
            // '+ 1' because calendar is 0-based for this field, and we are
            // 1-based
            t = -1;
            int tmon = mon;

            // get day...................................................
            boolean dayOfMSpec = daysOfMonth.contains(NO_SPEC) == false;
            boolean dayOfWSpec = daysOfWeek.contains(NO_SPEC) == false;
            if (dayOfMSpec && dayOfWSpec == false) { // get day by day of month rule
                st = daysOfMonth.tailSet(day);
                if (lastdayOfMonth) {
                    if (nearestWeekday == false) {
                        t = day;
                        day = getLastDayOfMonth(mon, cl.getYear());
                        day -= lastdayOffset;
                        if (t > day) {
                            mon++;
                            if (mon > 12) {
                                mon = 1;
                                tmon = 3333; // ensure test of mon != tmon further below fails
                                cl.plusYears(1);
                            }
                            day = 1;
                        }
                    } else {
                        t = day;
                        day = getLastDayOfMonth(mon, cl.getYear());
                        day -= lastdayOffset;

                        LocalDateTimeLegacyWrapper tcal = new LocalDateTimeLegacyWrapper(LocalDateTime.now(timeZone));
                        tcal.setSecond(0);
                        tcal.setMinute(0);
                        tcal.setHour(0);
                        tcal.setDayOfMonth(day);
                        tcal.setMonth(mon - 1);
                        tcal.setYear(cl.getYear());

                        int ldom = getLastDayOfMonth(mon, cl.getYear());
                        int dow = tcal.getDayOfWeek();

                        if (dow == Calendar.SATURDAY && day == 1) {
                            day += 2;
                        } else if (dow == Calendar.SATURDAY) {
                            day -= 1;
                        } else if (dow == Calendar.SUNDAY && day == ldom) {
                            day -= 2;
                        } else if (dow == Calendar.SUNDAY) {
                            day += 1;
                        }

                        tcal.setSecond(sec);
                        tcal.setMinute(min);
                        tcal.setHour(hr);
                        tcal.setDayOfMonth(day);
                        tcal.setMonth(mon - 1);
                        if (tcal.isBefore(afterTimeLdt)) {
                            day = 1;
                            mon++;
                        }
                    }
                } else if (nearestWeekday) {
                    t = day;
                    day = daysOfMonth.first();

                    LocalDateTimeLegacyWrapper tcal = new LocalDateTimeLegacyWrapper(LocalDateTime.now(timeZone));
                    tcal.setSecond(0);
                    tcal.setMinute(0);
                    tcal.setHour(0);
                    tcal.setDayOfMonth(day);
                    tcal.setMonth(mon - 1);
                    tcal.setYear(cl.getYear());

                    int ldom = getLastDayOfMonth(mon, cl.getYear());
                    int dow = tcal.getDayOfWeek();

                    if (dow == Calendar.SATURDAY && day == 1) {
                        day += 2;
                    } else if (dow == Calendar.SATURDAY) {
                        day -= 1;
                    } else if (dow == Calendar.SUNDAY && day == ldom) {
                        day -= 2;
                    } else if (dow == Calendar.SUNDAY) {
                        day += 1;
                    }

                    tcal.setSecond(sec);
                    tcal.setMinute(min);
                    tcal.setHour(hr);
                    tcal.setDayOfMonth(day);
                    tcal.setMonth(mon - 1);
                    if (tcal.isAfter(afterTimeLdt)) {
                        day = daysOfMonth.first();
                        mon++;
                    }
                } else if (st != null && st.size() != 0) {
                    t = day;
                    day = st.first();
                    // make sure we don't over-run a short month, such as february
                    int lastDay = getLastDayOfMonth(mon, cl.getYear());
                    if (day > lastDay) {
                        day = daysOfMonth.first();
                        mon++;
                    }
                } else {
                    day = daysOfMonth.first();
                    mon++;
                }

                if (day != t || mon != tmon) {
                    cl.setSecond(0);
                    cl.setMinute(0);
                    cl.setHour(0);
                    cl.setDayOfMonth(day);
                    cl.setMonth(mon - 1);
                    // '- 1' because calendar is 0-based for this field, and we
                    // are 1-based
                    continue;
                }
            } else if (dayOfWSpec && dayOfMSpec == false) { // get day by day of week rule
                if (lastdayOfWeek) { // are we looking for the last XXX day of
                    // the month?
                    int dow = daysOfWeek.first(); // desired
                    // d-o-w
                    int cDow = cl.getDayOfWeek(); // current d-o-w
                    int daysToAdd = 0;
                    if (cDow < dow) {
                        daysToAdd = dow - cDow;
                    }
                    if (cDow > dow) {
                        daysToAdd = dow + (7 - cDow);
                    }

                    int lDay = getLastDayOfMonth(mon, cl.getYear());

                    if (day + daysToAdd > lDay) { // did we already miss the
                        // last one?
                        cl.setSecond(0);
                        cl.setMinute(0);
                        cl.setHour(0);
                        cl.setDayOfMonth(1);
                        cl.setMonth(mon);
                        // no '- 1' here because we are promoting the month
                        continue;
                    }

                    // find date of last occurrence of this day in this month...
                    while ((day + daysToAdd + 7) <= lDay) {
                        daysToAdd += 7;
                    }

                    day += daysToAdd;

                    if (daysToAdd > 0) {
                        cl.setSecond(0);
                        cl.setMinute(0);
                        cl.setHour(0);
                        cl.setDayOfMonth(day);
                        cl.setMonth(mon - 1);
                        // '- 1' here because we are not promoting the month
                        continue;
                    }

                } else if (nthdayOfWeek != 0) {
                    // are we looking for the Nth XXX day in the month?
                    int dow = daysOfWeek.first(); // desired
                    // d-o-w
                    int cDow = cl.getDayOfWeek(); // current d-o-w
                    int daysToAdd = 0;
                    if (cDow < dow) {
                        daysToAdd = dow - cDow;
                    } else if (cDow > dow) {
                        daysToAdd = dow + (7 - cDow);
                    }

                    boolean dayShifted = false;
                    if (daysToAdd > 0) {
                        dayShifted = true;
                    }

                    day += daysToAdd;
                    int weekOfMonth = day / 7;
                    if (day % 7 > 0) {
                        weekOfMonth++;
                    }

                    daysToAdd = (nthdayOfWeek - weekOfMonth) * 7;
                    day += daysToAdd;
                    if (daysToAdd < 0 || day > getLastDayOfMonth(mon, cl.getYear())) {
                        cl.setSecond(0);
                        cl.setMinute(0);
                        cl.setHour(0);
                        cl.setDayOfMonth(1);
                        cl.setMonth(mon);
                        // no '- 1' here because we are promoting the month
                        continue;
                    } else if (daysToAdd > 0 || dayShifted) {
                        cl.setSecond(0);
                        cl.setMinute(0);
                        cl.setHour(0);
                        cl.setDayOfMonth(day);
                        cl.setMonth(mon - 1);
                        // '- 1' here because we are NOT promoting the month
                        continue;
                    }
                } else {
                    int cDow = cl.getDayOfWeek(); // current d-o-w
                    int dow = daysOfWeek.first(); // desired
                    // d-o-w
                    st = daysOfWeek.tailSet(cDow);
                    if (st != null && st.size() > 0) {
                        dow = st.first();
                    }

                    int daysToAdd = 0;
                    if (cDow < dow) {
                        daysToAdd = dow - cDow;
                    }
                    if (cDow > dow) {
                        daysToAdd = dow + (7 - cDow);
                    }

                    int lDay = getLastDayOfMonth(mon, cl.getYear());

                    if (day + daysToAdd > lDay) { // will we pass the end of
                        // the month?
                        cl.setSecond(0);
                        cl.setMinute(0);
                        cl.setHour(0);
                        cl.setDayOfMonth(1);
                        cl.setMonth(mon);
                        // no '- 1' here because we are promoting the month
                        continue;
                    } else if (daysToAdd > 0) { // are we swithing days?
                        cl.setSecond(0);
                        cl.setMinute(0);
                        cl.setHour(0);
                        cl.setDayOfMonth(day + daysToAdd);
                        cl.setMonth(mon - 1);
                        // '- 1' because calendar is 0-based for this field,
                        // and we are 1-based
                        continue;
                    }
                }
            } else { // dayOfWSpec && dayOfMSpec == false
                return -1;
                // throw new UnsupportedOperationException(
                // "Support for specifying both a day-of-week AND a day-of-month parameter is not implemented.");
            }
            cl.setDayOfMonth(day);

            mon = cl.getMonth() + 1;
            // '+ 1' because calendar is 0-based for this field, and we are
            // 1-based
            int year = cl.getYear();
            t = -1;

            // test for expressions that never generate a valid fire date,
            // but keep looping...
            if (year > MAX_YEAR) {
                return -1;
                // throw new ElasticsearchIllegalArgumentException("given time is not supported by cron [" + formatter.print(time) + "]");
            }

            // get month...................................................
            st = months.tailSet(mon);
            if (st != null && st.size() != 0) {
                t = mon;
                mon = st.first();
            } else {
                mon = months.first();
                year++;
            }
            if (mon != t) {
                cl.setSecond(0);
                cl.setMinute(0);
                cl.setHour(0);
                cl.setDayOfMonth(1);
                cl.setMonth(mon - 1);
                // '- 1' because calendar is 0-based for this field, and we are
                // 1-based
                cl.setYear(year);
                continue;
            }
            cl.setMonth(mon - 1);
            // '- 1' because calendar is 0-based for this field, and we are
            // 1-based

            year = cl.getYear();
            t = -1;

            // get year...................................................
            st = years.tailSet(year);
            if (st != null && st.size() != 0) {
                t = year;
                year = st.first();
            } else {
                return -1;
                // throw new ElasticsearchIllegalArgumentException("given time is not supported by cron [" + formatter.print(time) + "]");
            }

            if (year != t) {
                cl.setSecond(0);
                cl.setMinute(0);
                cl.setHour(0);
                cl.setDayOfMonth(1);
                cl.setMonth(0);
                // '- 1' because calendar is 0-based for this field, and we are
                // 1-based
                cl.setYear(year);
                continue;
            }
            cl.setYear(year);

            gotOne = true;
        } // while( done == false )

        LocalDateTime nextRuntime = cl.getLocalDateTime();

        return nextRuntime.atZone(timeZone).toInstant().toEpochMilli();
    }

    public String expression() {
        return expression;
    }

    public String getExpressionSummary() {
        StringBuilder buf = new StringBuilder();

        buf.append("seconds: ");
        buf.append(expressionSetSummary(seconds));
        buf.append("\n");
        buf.append("minutes: ");
        buf.append(expressionSetSummary(minutes));
        buf.append("\n");
        buf.append("hours: ");
        buf.append(expressionSetSummary(hours));
        buf.append("\n");
        buf.append("daysOfMonth: ");
        buf.append(expressionSetSummary(daysOfMonth));
        buf.append("\n");
        buf.append("months: ");
        buf.append(expressionSetSummary(months));
        buf.append("\n");
        buf.append("daysOfWeek: ");
        buf.append(expressionSetSummary(daysOfWeek));
        buf.append("\n");
        buf.append("lastdayOfWeek: ");
        buf.append(lastdayOfWeek);
        buf.append("\n");
        buf.append("nearestWeekday: ");
        buf.append(nearestWeekday);
        buf.append("\n");
        buf.append("NthDayOfWeek: ");
        buf.append(nthdayOfWeek);
        buf.append("\n");
        buf.append("lastdayOfMonth: ");
        buf.append(lastdayOfMonth);
        buf.append("\n");
        buf.append("years: ");
        buf.append(expressionSetSummary(years));
        buf.append("\n");

        return buf.toString();
    }

    @Override
    public int hashCode() {
        return Objects.hash(expression, timeZone);
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null || getClass() != obj.getClass()) {
            return false;
        }
        final Cron other = (Cron) obj;
        return Objects.equals(this.expression, other.expression) && Objects.equals(this.timeZone, other.timeZone);
    }

    /**
     * Returns the string representation of the <CODE>CronExpression</CODE>
     *
     * @return a string representation of the <CODE>CronExpression</CODE>
     */
    @Override
    public String toString() {
        return "Cron{" + "timeZone=" + timeZone + ", expression='" + expression + '\'' + '}';
    }

    /**
     * Indicates whether the specified cron expression can be parsed into a
     * valid cron expression
     *
     * @param expression the expression to evaluate
     * @return a boolean indicating whether the given expression is a valid cron
     *         expression
     */
    public static boolean isValid(String expression) {
        try {
            validate(expression);
        } catch (IllegalArgumentException pe) {
            return false;
        }
        return true;
    }

    public static void validate(String expression) throws IllegalArgumentException {
        new Cron(expression);
    }

    ////////////////////////////////////////////////////////////////////////////
    //
    // Expression Parsing Functions
    //
    ////////////////////////////////////////////////////////////////////////////

    private void buildExpression(@SuppressWarnings("HiddenField") String expression) {

        try {

            if (seconds == null) {
                seconds = new TreeSet<Integer>();
            }
            if (minutes == null) {
                minutes = new TreeSet<Integer>();
            }
            if (hours == null) {
                hours = new TreeSet<Integer>();
            }
            if (daysOfMonth == null) {
                daysOfMonth = new TreeSet<Integer>();
            }
            if (months == null) {
                months = new TreeSet<Integer>();
            }
            if (daysOfWeek == null) {
                daysOfWeek = new TreeSet<Integer>();
            }
            if (years == null) {
                years = new TreeSet<Integer>();
            }

            int exprOn = SECOND;

            StringTokenizer exprsTok = new StringTokenizer(expression, " \t", false);

            while (exprsTok.hasMoreTokens() && exprOn <= YEAR) {
                String expr = exprsTok.nextToken().trim();

                // throw an exception if L is used with other days of the month
                if (exprOn == DAY_OF_MONTH && expr.indexOf('L') != -1 && expr.length() > 1 && expr.contains(",")) {
                    throw illegalArgument("support for specifying 'L' and 'LW' with other days of the month is not implemented");
                }
                // throw an exception if L is used with other days of the week
                if (exprOn == DAY_OF_WEEK && expr.indexOf('L') != -1 && expr.length() > 1 && expr.contains(",")) {
                    throw illegalArgument("support for specifying 'L' with other days of the week is not implemented");
                }
                if (exprOn == DAY_OF_WEEK && expr.indexOf('#') != -1 && expr.indexOf('#', expr.indexOf('#') + 1) != -1) {
                    throw illegalArgument("support for specifying multiple \"nth\" days is not implemented.");
                }

                StringTokenizer vTok = new StringTokenizer(expr, ",");
                while (vTok.hasMoreTokens()) {
                    String v = vTok.nextToken();
                    storeExpressionVals(0, v, exprOn);
                }

                exprOn++;
            }

            if (exprOn <= DAY_OF_WEEK) {
                throw illegalArgument("unexpected end of expression at pos [{}].", expression.length());
            }

            if (exprOn <= YEAR) {
                storeExpressionVals(0, "*", YEAR);
            }

            TreeSet<Integer> dow = getSet(DAY_OF_WEEK);
            TreeSet<Integer> dom = getSet(DAY_OF_MONTH);

            // Copying the logic from the UnsupportedOperationException below
            boolean dayOfMSpec = dom.contains(NO_SPEC) == false;
            boolean dayOfWSpec = dow.contains(NO_SPEC) == false;

            if (dayOfMSpec == false || dayOfWSpec) {
                if (dayOfWSpec == false || dayOfMSpec) {
                    throw illegalArgument("support for specifying both a day-of-week AND a day-of-month parameter is not implemented.");
                }
            }
        } catch (Exception e) {
            throw illegalArgument("illegal cron expression format [{}]", e.toString());
        }
    }

    private int storeExpressionVals(int pos, String s, int type) throws ElasticsearchParseException {

        int incr = 0;
        int i = skipWhiteSpace(pos, s);
        if (i >= s.length()) {
            return i;
        }
        char c = s.charAt(i);
        if ((c >= 'A') && (c <= 'Z') && (s.equals("L") == false) && (s.equals("LW") == false) && (s.matches("^L-[0-9]*[W]?") == false)) {
            String sub = s.substring(i, i + 3);
            int sval = -1;
            int eval = -1;
            if (type == MONTH) {
                sval = getMonthNumber(sub) + 1;
                if (sval <= 0) {
                    throw illegalArgument("invalid Month value [{}] at pos [{}]", sub, i);
                }
                if (s.length() > i + 3) {
                    c = s.charAt(i + 3);
                    if (c == '-') {
                        i += 4;
                        sub = s.substring(i, i + 3);
                        eval = getMonthNumber(sub) + 1;
                        if (eval <= 0) {
                            throw illegalArgument("invalid Month value [{}] at pos [{}]", sub, i);
                        }
                    }
                }
            } else if (type == DAY_OF_WEEK) {
                sval = getDayOfWeekNumber(sub);
                if (sval < 0) {
                    throw illegalArgument("invalid Day-of-Week value [{}] at pos [{}]", sub, i);
                }
                if (s.length() > i + 3) {
                    c = s.charAt(i + 3);
                    if (c == '-') {
                        i += 4;
                        sub = s.substring(i, i + 3);
                        eval = getDayOfWeekNumber(sub);
                        if (eval < 0) {
                            throw illegalArgument("invalid Day-of-Week value [{}] at pos [{}]", sub, i);
                        }
                    } else if (c == '#') {
                        try {
                            i += 4;
                            nthdayOfWeek = Integer.parseInt(s.substring(i));
                            if (nthdayOfWeek < 1 || nthdayOfWeek > 5) {
                                throw new Exception();
                            }
                        } catch (Exception e) {
                            throw illegalArgument("a numeric value between 1 and 5 must follow the '#' option at pos [{}]", i);
                        }
                    } else if (c == 'L') {
                        lastdayOfWeek = true;
                        i++;
                    }
                }

            } else {
                throw illegalArgument("illegal characters [{}] at pos [{}] '", sub, i);
            }
            if (eval != -1) {
                incr = 1;
            }
            addToSet(sval, eval, incr, type);
            return (i + 3);
        }

        if (c == '?') {
            i++;
            if ((i + 1) < s.length() && (s.charAt(i) != ' ' && s.charAt(i + 1) != '\t')) {
                throw illegalArgument("illegal character [{}] after '?' at pos [{}]", s.charAt(i), i);
            }
            if (type != DAY_OF_WEEK && type != DAY_OF_MONTH) {
                throw illegalArgument("'?' can only be specified for Day-of-Month or Day-of-Week. at pos [{}]", i);
            }
            if (type == DAY_OF_WEEK && lastdayOfMonth == false) {
                int val = daysOfMonth.last();
                if (val == NO_SPEC_INT) {
                    throw illegalArgument("'?' can only be specified for Day-of-Month -OR- Day-of-Week. at pos [{}]", i);
                }
            }

            addToSet(NO_SPEC_INT, -1, 0, type);
            return i;
        }

        if (c == '*' || c == '/') {
            if (c == '*' && (i + 1) >= s.length()) {
                addToSet(ALL_SPEC_INT, -1, incr, type);
                return i + 1;
            } else if (c == '/' && ((i + 1) >= s.length() || s.charAt(i + 1) == ' ' || s.charAt(i + 1) == '\t')) {
                throw illegalArgument("'/' must be followed by an integer. at pos [{}]", i);
            } else if (c == '*') {
                i++;
            }
            c = s.charAt(i);
            if (c == '/') { // is an increment specified?
                i++;
                if (i >= s.length()) {
                    throw illegalArgument("Unexpected end of string. at pos [{}]", i);
                }

                incr = getNumericValue(s, i);

                i++;
                if (incr > 10) {
                    i++;
                }
                if (incr > 59 && (type == SECOND || type == MINUTE)) {
                    throw illegalArgument("increment [{}] > 60 at pos [{}]", incr, i);
                } else if (incr > 23 && (type == HOUR)) {
                    throw illegalArgument("increment [{}] > 24 at pos [{}]", incr, i);
                } else if (incr > 31 && (type == DAY_OF_MONTH)) {
                    throw illegalArgument("increment [{}] > 31 at pos [{}] ", incr, i);
                } else if (incr > 7 && (type == DAY_OF_WEEK)) {
                    throw illegalArgument("increment [{}] > 7 at pos [{}] ", incr, i);
                } else if (incr > 12 && (type == MONTH)) {
                    throw illegalArgument("increment [{}] > 12 at pos [{}]", incr, i);
                }
            } else {
                incr = 1;
            }

            addToSet(ALL_SPEC_INT, -1, incr, type);
            return i;
        } else if (c == 'L') {
            i++;
            if (type == DAY_OF_MONTH) {
                lastdayOfMonth = true;
            }
            if (type == DAY_OF_WEEK) {
                addToSet(7, 7, 0, type);
            }
            if (type == DAY_OF_MONTH && s.length() > i) {
                c = s.charAt(i);
                if (c == '-') {
                    ValueSet vs = getValue(0, s, i + 1);
                    lastdayOffset = vs.value;
                    if (lastdayOffset > 30) throw illegalArgument("offset from last day must be <= 30 at pos [{}]", i + 1);
                    i = vs.pos;
                }
                if (s.length() > i) {
                    c = s.charAt(i);
                    if (c == 'W') {
                        nearestWeekday = true;
                        i++;
                    }
                }
            }
            return i;
        } else if (c >= '0' && c <= '9') {
            int val = Integer.parseInt(String.valueOf(c));
            i++;
            if (i >= s.length()) {
                addToSet(val, -1, -1, type);
            } else {
                c = s.charAt(i);
                if (c >= '0' && c <= '9') {
                    ValueSet vs = getValue(val, s, i);
                    val = vs.value;
                    i = vs.pos;
                }
                i = checkNext(i, s, val, type);
                return i;
            }
        } else {
            throw illegalArgument("Unexpected character [{}] at pos [{}] ", c, i);
        }

        return i;
    }

    private int checkNext(int pos, String s, int val, int type) throws ElasticsearchParseException {

        int end = -1;
        int i = pos;

        if (i >= s.length()) {
            addToSet(val, end, -1, type);
            return i;
        }

        char c = s.charAt(pos);

        if (c == 'L') {
            if (type == DAY_OF_WEEK) {
                if (val < 1 || val > 7) throw illegalArgument("Day-of-Week values must be between 1 and 7");
                lastdayOfWeek = true;
            } else {
                throw illegalArgument("'L' option is not valid here. at pos [{}]", i);
            }
            TreeSet<Integer> set = getSet(type);
            set.add(val);
            i++;
            return i;
        }

        if (c == 'W') {
            if (type == DAY_OF_MONTH) {
                nearestWeekday = true;
            } else {
                throw illegalArgument("'W' option is not valid here. at pos [{}]", i);
            }
            if (val > 31) throw illegalArgument(
                "the 'W' option does not make sense with values larger than 31 (max number of days in a month) at " + "pos [{}]",
                i
            );
            TreeSet<Integer> set = getSet(type);
            set.add(val);
            i++;
            return i;
        }

        if (c == '#') {
            if (type != DAY_OF_WEEK) {
                throw illegalArgument("'#' option is not valid here. at pos [{}]", i);
            }
            i++;
            try {
                nthdayOfWeek = Integer.parseInt(s.substring(i));
                if (nthdayOfWeek < 1 || nthdayOfWeek > 5) {
                    throw new Exception();
                }
            } catch (Exception e) {
                throw illegalArgument("a numeric value between 1 and 5 must follow the '#' option at pos [{}]", i);
            }

            TreeSet<Integer> set = getSet(type);
            set.add(val);
            i++;
            return i;
        }

        if (c == '-') {
            i++;
            c = s.charAt(i);
            int v = Integer.parseInt(String.valueOf(c));
            end = v;
            i++;
            if (i >= s.length()) {
                addToSet(val, end, 1, type);
                return i;
            }
            c = s.charAt(i);
            if (c >= '0' && c <= '9') {
                ValueSet vs = getValue(v, s, i);
                end = vs.value;
                i = vs.pos;
            }
            if (i < s.length() && ((c = s.charAt(i)) == '/')) {
                i++;
                c = s.charAt(i);
                int v2 = Integer.parseInt(String.valueOf(c));
                i++;
                if (i >= s.length()) {
                    addToSet(val, end, v2, type);
                    return i;
                }
                c = s.charAt(i);
                if (c >= '0' && c <= '9') {
                    ValueSet vs = getValue(v2, s, i);
                    int v3 = vs.value;
                    addToSet(val, end, v3, type);
                    i = vs.pos;
                    return i;
                } else {
                    addToSet(val, end, v2, type);
                    return i;
                }
            } else {
                addToSet(val, end, 1, type);
                return i;
            }
        }

        if (c == '/') {
            i++;
            c = s.charAt(i);
            int v2 = Integer.parseInt(String.valueOf(c));
            i++;
            if (i >= s.length()) {
                addToSet(val, end, v2, type);
                return i;
            }
            c = s.charAt(i);
            if (c >= '0' && c <= '9') {
                ValueSet vs = getValue(v2, s, i);
                int v3 = vs.value;
                addToSet(val, end, v3, type);
                i = vs.pos;
                return i;
            } else {
                throw illegalArgument("Unexpected character [{}] after '/' at pos [{}]", c, i);
            }
        }

        addToSet(val, end, 0, type);
        i++;
        return i;
    }

    private static String expressionSetSummary(java.util.Set<Integer> set) {

        if (set.contains(NO_SPEC)) {
            return "?";
        }
        if (set.contains(ALL_SPEC)) {
            return "*";
        }

        StringBuilder buf = new StringBuilder();

        Iterator<Integer> itr = set.iterator();
        boolean first = true;
        while (itr.hasNext()) {
            Integer iVal = itr.next();
            String val = iVal.toString();
            if (first == false) {
                buf.append(",");
            }
            buf.append(val);
            first = false;
        }

        return buf.toString();
    }

    private static int skipWhiteSpace(int i, String s) {
        for (; i < s.length() && (s.charAt(i) == ' ' || s.charAt(i) == '\t'); i++) {
            // intentionally empty
        }

        return i;
    }

    private static int findNextWhiteSpace(int i, String s) {
        for (; i < s.length() && (s.charAt(i) != ' ' || s.charAt(i) != '\t'); i++) {
            // intentionally empty
        }

        return i;
    }

    private void addToSet(int val, int end, int incr, int type) throws ElasticsearchParseException {

        TreeSet<Integer> set = getSet(type);

        if (type == SECOND || type == MINUTE) {
            if ((val < 0 || val > 59 || end > 59) && (val != ALL_SPEC_INT)) {
                throw illegalArgument("Minute and Second values must be between 0 and 59");
            }
        } else if (type == HOUR) {
            if ((val < 0 || val > 23 || end > 23) && (val != ALL_SPEC_INT)) {
                throw illegalArgument("Hour values must be between 0 and 23");
            }
        } else if (type == DAY_OF_MONTH) {
            if ((val < 1 || val > 31 || end > 31) && (val != ALL_SPEC_INT) && (val != NO_SPEC_INT)) {
                throw illegalArgument("Day of month values must be between 1 and 31");
            }
        } else if (type == MONTH) {
            if ((val < 1 || val > 12 || end > 12) && (val != ALL_SPEC_INT)) {
                throw illegalArgument("Month values must be between 1 and 12");
            }
        } else if (type == DAY_OF_WEEK) {
            if ((val == 0 || val > 7 || end > 7) && (val != ALL_SPEC_INT) && (val != NO_SPEC_INT)) {
                throw illegalArgument("Day-of-Week values must be between 1 and 7");
            }
        }

        if ((incr == 0 || incr == -1) && val != ALL_SPEC_INT) {
            if (val != -1) {
                set.add(val);
            } else {
                set.add(NO_SPEC);
            }

            return;
        }

        int startAt = val;
        int stopAt = end;

        if (val == ALL_SPEC_INT && incr <= 0) {
            incr = 1;
            set.add(ALL_SPEC); // put in a marker, but also fill values
        }

        if (type == SECOND || type == MINUTE) {
            if (stopAt == -1) {
                stopAt = 59;
            }
            if (startAt == -1 || startAt == ALL_SPEC_INT) {
                startAt = 0;
            }
        } else if (type == HOUR) {
            if (stopAt == -1) {
                stopAt = 23;
            }
            if (startAt == -1 || startAt == ALL_SPEC_INT) {
                startAt = 0;
            }
        } else if (type == DAY_OF_MONTH) {
            if (stopAt == -1) {
                stopAt = 31;
            }
            if (startAt == -1 || startAt == ALL_SPEC_INT) {
                startAt = 1;
            }
        } else if (type == MONTH) {
            if (stopAt == -1) {
                stopAt = 12;
            }
            if (startAt == -1 || startAt == ALL_SPEC_INT) {
                startAt = 1;
            }
        } else if (type == DAY_OF_WEEK) {
            if (stopAt == -1) {
                stopAt = 7;
            }
            if (startAt == -1 || startAt == ALL_SPEC_INT) {
                startAt = 1;
            }
        } else if (type == YEAR) {
            if (stopAt == -1) {
                stopAt = MAX_YEAR;
            }
            if (startAt == -1 || startAt == ALL_SPEC_INT) {
                startAt = 1970;
            }
        }

        // if the end of the range is before the start, then we need to overflow into
        // the next day, month etc. This is done by adding the maximum amount for that
        // type, and using modulus max to determine the value being added.
        int max = -1;
        if (stopAt < startAt) {
            max = switch (type) {
                case SECOND -> 60;
                case MINUTE -> 60;
                case HOUR -> 24;
                case MONTH -> 12;
                case DAY_OF_WEEK -> 7;
                case DAY_OF_MONTH -> 31;
                case YEAR -> throw new IllegalArgumentException("Start year must be less than stop year");
                default -> throw new IllegalArgumentException("Unexpected type encountered");
            };
            stopAt += max;
        }

        for (int i = startAt; i <= stopAt; i += incr) {
            if (max == -1) {
                // ie: there's no max to overflow over
                set.add(i);
            } else {
                // take the modulus to get the real value
                int i2 = i % max;

                // 1-indexed ranges should not include 0, and should include their max
                if (i2 == 0 && (type == MONTH || type == DAY_OF_WEEK || type == DAY_OF_MONTH)) {
                    i2 = max;
                }

                set.add(i2);
            }
        }
    }

    private TreeSet<Integer> getSet(int type) {
        return switch (type) {
            case SECOND -> seconds;
            case MINUTE -> minutes;
            case HOUR -> hours;
            case DAY_OF_MONTH -> daysOfMonth;
            case MONTH -> months;
            case DAY_OF_WEEK -> daysOfWeek;
            case YEAR -> years;
            default -> null;
        };
    }

    private static ValueSet getValue(int v, String s, int i) {
        char c = s.charAt(i);
        StringBuilder s1 = new StringBuilder(String.valueOf(v));
        while (c >= '0' && c <= '9') {
            s1.append(c);
            i++;
            if (i >= s.length()) {
                break;
            }
            c = s.charAt(i);
        }
        ValueSet val = new ValueSet();

        val.pos = (i < s.length()) ? i : i + 1;
        val.value = Integer.parseInt(s1.toString());
        return val;
    }

    private static int getNumericValue(String s, int i) {
        int endOfVal = findNextWhiteSpace(i, s);
        String val = s.substring(i, endOfVal);
        return Integer.parseInt(val);
    }

    private static int getMonthNumber(String s) {
        Integer integer = MONTH_MAP.get(s);

        if (integer == null) {
            return -1;
        }

        return integer;
    }

    private static int getDayOfWeekNumber(String s) {
        Integer integer = DAY_MAP.get(s);

        if (integer == null) {
            return -1;
        }

        return integer;
    }

    /**
     * Advance the calendar to the particular hour paying particular attention
     * to daylight saving problems.
     *
     * @param cal the calendar to operate on
     * @param hour the hour to set
     */
    private static void setCalendarHour(Calendar cal, int hour) {
        cal.set(java.util.Calendar.HOUR_OF_DAY, hour);
        if (cal.get(java.util.Calendar.HOUR_OF_DAY) != hour && hour != 24) {
            cal.set(java.util.Calendar.HOUR_OF_DAY, hour + 1);
        }
    }

    private static boolean isLeapYear(int year) {
        return ((year % 4 == 0 && year % 100 != 0) || (year % 400 == 0));
    }

    private static int getLastDayOfMonth(int monthNum, int year) {

        return switch (monthNum) {
            case 1 -> 31;
            case 2 -> (isLeapYear(year)) ? 29 : 28;
            case 3 -> 31;
            case 4 -> 30;
            case 5 -> 31;
            case 6 -> 30;
            case 7 -> 31;
            case 8 -> 31;
            case 9 -> 30;
            case 10 -> 31;
            case 11 -> 30;
            case 12 -> 31;
            default -> throw new IllegalArgumentException("Illegal month number: " + monthNum);
        };
    }

    @Override
    public XContentBuilder toXContent(XContentBuilder builder, Params params) throws IOException {
        return builder.value(expression);
    }

    private static class ValueSet {
        int value;
        int pos;
    }
}
