/*
 * Copyright Elasticsearch B.V. and/or licensed to Elasticsearch B.V. under one
 * or more contributor license agreements. Licensed under the Elastic License
 * 2.0; you may not use this file except in compliance with the Elastic License
 * 2.0.
 */

package org.elasticsearch.xpack.ilm;

import org.apache.http.entity.ContentType;
import org.apache.http.entity.StringEntity;
import org.apache.http.util.EntityUtils;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.elasticsearch.client.Request;
import org.elasticsearch.client.Response;
import org.elasticsearch.cluster.metadata.IndexMetadata;
import org.elasticsearch.common.Strings;
import org.elasticsearch.common.settings.Settings;
import org.elasticsearch.core.TimeValue;
import org.elasticsearch.test.rest.ESRestTestCase;
import org.elasticsearch.xcontent.XContentBuilder;
import org.elasticsearch.xpack.core.ilm.DeleteAction;
import org.elasticsearch.xpack.core.ilm.ErrorStep;
import org.elasticsearch.xpack.core.ilm.LifecycleAction;
import org.elasticsearch.xpack.core.ilm.LifecyclePolicy;
import org.elasticsearch.xpack.core.ilm.LifecycleSettings;
import org.elasticsearch.xpack.core.ilm.Phase;
import org.elasticsearch.xpack.core.ilm.RolloverAction;
import org.elasticsearch.xpack.core.ilm.ShrinkAction;
import org.junit.Before;

import java.util.Formatter;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;
import java.util.concurrent.TimeUnit;

import static org.elasticsearch.xcontent.XContentFactory.jsonBuilder;
import static org.elasticsearch.xpack.TimeSeriesRestDriver.createFullPolicy;
import static org.elasticsearch.xpack.TimeSeriesRestDriver.createIndexWithSettings;
import static org.elasticsearch.xpack.TimeSeriesRestDriver.createNewSingletonPolicy;
import static org.elasticsearch.xpack.TimeSeriesRestDriver.explain;
import static org.elasticsearch.xpack.TimeSeriesRestDriver.explainIndex;
import static org.hamcrest.Matchers.allOf;
import static org.hamcrest.Matchers.containsString;
import static org.hamcrest.Matchers.greaterThan;
import static org.hamcrest.Matchers.greaterThanOrEqualTo;
import static org.hamcrest.Matchers.hasKey;
import static org.hamcrest.Matchers.is;
import static org.hamcrest.Matchers.not;
import static org.hamcrest.Matchers.notNullValue;
import static org.hamcrest.Matchers.nullValue;

public class ExplainLifecycleIT extends ESRestTestCase {
    private static final Logger logger = LogManager.getLogger(ExplainLifecycleIT.class);
    private static final String FAILED_STEP_RETRY_COUNT_FIELD = "failed_step_retry_count";
    private static final String IS_AUTO_RETRYABLE_ERROR_FIELD = "is_auto_retryable_error";

    private String policy;
    private String index;
    private String alias;

    @Before
    public void refreshIndex() {
        index = "index-" + randomAlphaOfLength(10).toLowerCase(Locale.ROOT);
        policy = "policy-" + randomAlphaOfLength(5);
        alias = "alias-" + randomAlphaOfLength(5);
    }

    public void testExplainFilters() throws Exception {
        String goodIndex = index + "-good-000001";
        String errorIndex = index + "-error";
        String nonexistantPolicyIndex = index + "-nonexistant-policy";
        String unmanagedIndex = index + "-unmanaged";

        createFullPolicy(client(), policy, TimeValue.ZERO);

        {
            // Create a "shrink-only-policy"
            Map<String, LifecycleAction> warmActions = new HashMap<>();
            warmActions.put(ShrinkAction.NAME, new ShrinkAction(17, null, false));
            Map<String, Phase> phases = new HashMap<>();
            phases.put("warm", new Phase("warm", TimeValue.ZERO, warmActions));
            LifecyclePolicy lifecyclePolicy = new LifecyclePolicy("shrink-only-policy", phases);
            // PUT policy
            XContentBuilder builder = jsonBuilder();
            lifecyclePolicy.toXContent(builder, null);
            final StringEntity entity = new StringEntity("{ \"policy\":" + Strings.toString(builder) + "}", ContentType.APPLICATION_JSON);
            Request request = new Request("PUT", "_ilm/policy/shrink-only-policy");
            request.setEntity(entity);
            assertOK(client().performRequest(request));
        }

        createIndexWithSettings(
            client(),
            goodIndex,
            alias,
            Settings.builder()
                .put(RolloverAction.LIFECYCLE_ROLLOVER_ALIAS, alias)
                .put(IndexMetadata.SETTING_NUMBER_OF_REPLICAS, 0)
                .put(LifecycleSettings.LIFECYCLE_NAME, policy)
        );
        createIndexWithSettings(
            client(),
            errorIndex,
            Settings.builder().put(IndexMetadata.SETTING_NUMBER_OF_REPLICAS, 0).put(LifecycleSettings.LIFECYCLE_NAME, "shrink-only-policy")
        );
        createIndexWithSettings(
            client(),
            nonexistantPolicyIndex,
            Settings.builder()
                .put(IndexMetadata.SETTING_NUMBER_OF_REPLICAS, 0)
                .put(LifecycleSettings.LIFECYCLE_NAME, randomValueOtherThan(policy, () -> randomAlphaOfLengthBetween(3, 10)))
        );
        createIndexWithSettings(client(), unmanagedIndex, Settings.builder().put(IndexMetadata.SETTING_NUMBER_OF_REPLICAS, 0));

        assertBusy(() -> {
            Map<String, Map<String, Object>> explainResponse = explain(client(), index + "*", false, false);
            assertNotNull(explainResponse);
            assertThat(
                explainResponse,
                allOf(hasKey(goodIndex), hasKey(errorIndex), hasKey(nonexistantPolicyIndex), hasKey(unmanagedIndex))
            );

            Map<String, Map<String, Object>> onlyManagedResponse = explain(client(), index + "*", false, true);
            assertNotNull(onlyManagedResponse);
            assertThat(onlyManagedResponse, allOf(hasKey(goodIndex), hasKey(errorIndex), hasKey(nonexistantPolicyIndex)));
            assertThat(onlyManagedResponse, not(hasKey(unmanagedIndex)));

            Map<String, Map<String, Object>> onlyErrorsResponse = explain(client(), index + "*", true, true);
            assertNotNull(onlyErrorsResponse);
            assertThat(onlyErrorsResponse, hasKey(nonexistantPolicyIndex));
            assertThat(onlyErrorsResponse, allOf(not(hasKey(goodIndex)), not(hasKey(unmanagedIndex))));
        });
    }

    public void testExplainIndexContainsAutomaticRetriesInformation() throws Exception {
        createFullPolicy(client(), policy, TimeValue.ZERO);

        // create index without alias so the rollover action fails and is retried
        createIndexWithSettings(
            client(),
            index,
            Settings.builder().put(IndexMetadata.SETTING_NUMBER_OF_REPLICAS, 0).put(LifecycleSettings.LIFECYCLE_NAME, policy)
        );

        assertBusy(() -> {
            Map<String, Object> explainIndex = explainIndex(client(), index);
            assertThat((Integer) explainIndex.get(FAILED_STEP_RETRY_COUNT_FIELD), greaterThanOrEqualTo(1));
            assertThat(explainIndex.get(IS_AUTO_RETRYABLE_ERROR_FIELD), is(true));
        });
    }

    @SuppressWarnings("unchecked")
    public void testExplainIndicesWildcard() throws Exception {
        createNewSingletonPolicy(client(), policy, "delete", DeleteAction.WITH_SNAPSHOT_DELETE, TimeValue.timeValueDays(100));
        String firstIndex = this.index + "-first";
        String secondIndex = this.index + "-second";
        String unmanagedIndex = this.index + "-unmanaged";
        String indexWithMissingPolicy = this.index + "-missing_policy";
        createIndexWithSettings(
            client(),
            firstIndex,
            alias + firstIndex,
            Settings.builder()
                .put(IndexMetadata.SETTING_NUMBER_OF_SHARDS, 1)
                .put(IndexMetadata.SETTING_NUMBER_OF_REPLICAS, 0)
                .put(LifecycleSettings.LIFECYCLE_NAME, policy)
        );
        createIndexWithSettings(
            client(),
            secondIndex,
            alias + secondIndex,
            Settings.builder()
                .put(IndexMetadata.SETTING_NUMBER_OF_SHARDS, 1)
                .put(IndexMetadata.SETTING_NUMBER_OF_REPLICAS, 0)
                .put(LifecycleSettings.LIFECYCLE_NAME, policy)
        );
        createIndexWithSettings(
            client(),
            unmanagedIndex,
            alias + unmanagedIndex,
            Settings.builder().put(IndexMetadata.SETTING_NUMBER_OF_SHARDS, 1).put(IndexMetadata.SETTING_NUMBER_OF_REPLICAS, 0)
        );
        String missingPolicyName = "missing_policy_";
        createIndexWithSettings(
            client(),
            indexWithMissingPolicy,
            alias + indexWithMissingPolicy,
            Settings.builder()
                .put(IndexMetadata.SETTING_NUMBER_OF_SHARDS, 1)
                .put(IndexMetadata.SETTING_NUMBER_OF_REPLICAS, 0)
                .put(LifecycleSettings.LIFECYCLE_NAME, missingPolicyName)
        );

        assertBusy(() -> {
            Map<String, Map<String, Object>> explain = explain(client(), this.index + "*", false, false);
            assertManagedIndex(explain.get(firstIndex));
            assertManagedIndex(explain.get(secondIndex));
            assertUnmanagedIndex(explain.get(unmanagedIndex));

            Map<String, Object> explainIndexWithMissingPolicy = explain.get(indexWithMissingPolicy);
            assertThat(explainIndexWithMissingPolicy.get("managed"), is(true));
            assertThat(explainIndexWithMissingPolicy.get("policy"), is(missingPolicyName));
            assertThat(explainIndexWithMissingPolicy.get("phase"), is(nullValue()));
            assertThat(explainIndexWithMissingPolicy.get("action"), is(nullValue()));
            assertThat(explainIndexWithMissingPolicy.get("step"), is(ErrorStep.NAME));
            assertThat(explainIndexWithMissingPolicy.get("age"), is(nullValue()));
            assertThat(explainIndexWithMissingPolicy.get("failed_step"), is(nullValue()));
            Map<String, Object> stepInfo = (Map<String, Object>) explainIndexWithMissingPolicy.get("step_info");
            assertThat(stepInfo, is(notNullValue()));
            assertThat(stepInfo.get("reason"), is("policy [missing_policy_] does not exist"));
        });
    }

    public void testExplainOrder() throws Exception {
        createNewSingletonPolicy(client(), policy, "delete", DeleteAction.WITH_SNAPSHOT_DELETE, TimeValue.timeValueDays(100));
        createIndexWithSettings(
            client(),
            "order-aaa-foo",
            "alias-1",
            Settings.builder()
                .put(IndexMetadata.SETTING_NUMBER_OF_SHARDS, 1)
                .put(IndexMetadata.SETTING_NUMBER_OF_REPLICAS, 0)
                .put(LifecycleSettings.LIFECYCLE_NAME, policy)
        );
        createIndexWithSettings(
            client(),
            "order-ccc-foo",
            "alias-2",
            Settings.builder()
                .put(IndexMetadata.SETTING_NUMBER_OF_SHARDS, 1)
                .put(IndexMetadata.SETTING_NUMBER_OF_REPLICAS, 0)
                .put(LifecycleSettings.LIFECYCLE_NAME, policy)
        );
        createIndexWithSettings(
            client(),
            "order-bbb-foo",
            "alias-3",
            Settings.builder()
                .put(IndexMetadata.SETTING_NUMBER_OF_SHARDS, 1)
                .put(IndexMetadata.SETTING_NUMBER_OF_REPLICAS, 0)
                .put(LifecycleSettings.LIFECYCLE_NAME, policy)
        );
        ensureGreen("order-*");

        Request req = new Request("GET", "/order-*/_ilm/explain?filter_path=indices.*.index");
        Response resp = client().performRequest(req);
        String respString = EntityUtils.toString(resp.getEntity());
        // Assert that indices are ordered based on a natural ordering of their index name
        assertThat(
            respString,
            containsString(
                "{\"indices\":{"
                    + "\"order-aaa-foo\":{\"index\":\"order-aaa-foo\"},"
                    + "\"order-bbb-foo\":{\"index\":\"order-bbb-foo\"},"
                    + "\"order-ccc-foo\":{\"index\":\"order-ccc-foo\"}}}"
            )
        );
    }

    public void testStepInfoPreservedOnAutoRetry() throws Exception {
        String policyName = "policy-" + randomAlphaOfLength(5).toLowerCase(Locale.ROOT);

        Request createPolice = new Request("PUT", "_ilm/policy/" + policyName);
        createPolice.setJsonEntity("""
            {
              "policy": {
                "phases": {
                  "hot": {
                    "actions": {
                      "rollover": {
                        "max_docs": 1
                      }
                    }
                  }
                }
              }
            }
            """);
        assertOK(client().performRequest(createPolice));

        String aliasName = "step-info-test";
        String indexName = aliasName + "-" + randomAlphaOfLength(5).toLowerCase(Locale.ROOT);

        Request templateRequest = new Request("PUT", "_index_template/template_" + policyName);

        String templateBodyTemplate = """
            {
              "index_patterns": ["%s-*"],
              "template": {
                "settings": {
                  "index.lifecycle.name": "%s",
                  "index.lifecycle.rollover_alias": "%s"
                }
              }
            }
            """;
        Formatter formatter = new Formatter(Locale.ROOT);
        templateRequest.setJsonEntity(formatter.format(templateBodyTemplate, aliasName, policyName, aliasName).toString());

        assertOK(client().performRequest(templateRequest));

        Request indexRequest = new Request("POST", "/" + indexName + "/_doc/1");
        indexRequest.setJsonEntity("{\"test\":\"value\"}");
        assertOK(client().performRequest(indexRequest));

        assertBusy(() -> {
            Map<String, Object> explainIndex = explainIndex(client(), indexName);
            var assertionMessage = "Assertion failed for the following response: " + explainIndex;
            assertThat(assertionMessage, explainIndex.get("failed_step_retry_count"), notNullValue());
            assertThat(assertionMessage, explainIndex.get("previous_step_info"), notNullValue());
            assertThat(assertionMessage, (int) explainIndex.get("failed_step_retry_count"), greaterThan(0));
            assertThat(
                assertionMessage,
                explainIndex.get("previous_step_info").toString(),
                containsString("rollover_alias [" + aliasName + "] does not point to index [" + indexName + "]")
            );
        }, 30, TimeUnit.SECONDS);
    }

    private void assertUnmanagedIndex(Map<String, Object> explainIndexMap) {
        assertThat(explainIndexMap.get("managed"), is(false));
        assertThat(explainIndexMap.get("time_since_index_creation"), is(nullValue()));
        assertThat(explainIndexMap.get("index_creation_date_millis"), is(nullValue()));
        assertThat(explainIndexMap.get("policy"), is(nullValue()));
        assertThat(explainIndexMap.get("phase"), is(nullValue()));
        assertThat(explainIndexMap.get("action"), is(nullValue()));
        assertThat(explainIndexMap.get("step"), is(nullValue()));
        assertThat(explainIndexMap.get("age"), is(nullValue()));
        assertThat(explainIndexMap.get("failed_step"), is(nullValue()));
        assertThat(explainIndexMap.get("step_info"), is(nullValue()));
    }

    private void assertManagedIndex(Map<String, Object> explainIndexMap) {
        assertThat(explainIndexMap.get("managed"), is(true));
        assertThat(explainIndexMap.get("time_since_index_creation"), is(notNullValue()));
        assertThat(explainIndexMap.get("index_creation_date_millis"), is(notNullValue()));
        assertThat(explainIndexMap.get("policy"), is(policy));
        assertThat(explainIndexMap.get("phase"), is("new"));
        assertThat(explainIndexMap.get("action"), is("complete"));
        assertThat(explainIndexMap.get("step"), is("complete"));
        assertThat(explainIndexMap.get("phase_time_millis"), is(notNullValue()));
        assertThat(explainIndexMap.get("age"), is(notNullValue()));
        assertThat(explainIndexMap.get("phase_execution"), is(notNullValue()));
        assertThat(explainIndexMap.get("failed_step"), is(nullValue()));
        assertThat(explainIndexMap.get("step_info"), is(nullValue()));
    }

}
